# (C) British Crown Copyright 2019, Met Office
#
# This file is part of cf-units.
#
# cf-units is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cf-units is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cf-units.  If not, see <http://www.gnu.org/licenses/>.

import six

if six.PY2:
    raise ImportError('Python3 required for cf-units latex support')

import cf_units._udunits2_parser.graph as graph  # noqa: E402
from cf_units._udunits2_parser import parse as _parse  # noqa: E402


class TeXVisitor(graph.Visitor):
    def _format(self, fmt, lhs, rhs):
        return fmt.format(self.visit(lhs), self.visit(rhs))

    def visit_Identifier(self, node):
        token = str(node)
        if token.startswith('micro'):
            token = token.replace('micro', '{\mu}')
        return token

    def visit_Raise(self, node):
        return self._format('{{{}}}^{{{}}}', node.lhs, node.rhs)

    def visit_Multiply(self, node):
        return self._format(r'{{{}}}\cdot{{{}}}', node.lhs, node.rhs)

    def visit_Divide(self, node):
        return self._format(r'\frac{{{}}}{{{}}}', node.lhs, node.rhs)

    def visit_Shift(self, node):
        return self._format('{{{}}} @ {{{}}}', node.unit, node.shift_from)

    def generic_visit(self, node):
        result = [self.visit(child) for child in node.children()]
        if not result:
            result = str(node)
        return result


def tex(unit_str):
    tree = _parse(unit_str)
    return TeXVisitor().visit(tree)



# (C) British Crown Copyright 2016 - 2019, Met Office
#
# This file is part of cf-units.
#
# cf-units is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cf-units is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cf-units.  If not, see <http://www.gnu.org/licenses/>.
"""Test function :func:`cf_units.date2num`."""

from __future__ import (absolute_import, division, print_function)
from six.moves import (filter, input, map, range, zip)  # noqa

import unittest
import datetime

import numpy as np
import six

from cf_units import date2num


class Test(unittest.TestCase):
    def setUp(self):
        self.unit = 'seconds since 1970-01-01'
        self.calendar = 'gregorian'

    def test_single(self):
        date = datetime.datetime(1970, 1, 1, 0, 0, 5)
        exp = 5.
        res = date2num(date, self.unit, self.calendar)
        # num2date won't return an exact value representing the date,
        # even if one exists
        self.assertAlmostEqual(exp, res, places=4)

    def test_sequence(self):
        dates = [datetime.datetime(1970, 1, 1, 0, 0, 20),
                 datetime.datetime(1970, 1, 1, 0, 0, 40),
                 datetime.datetime(1970, 1, 1, 0, 1),
                 datetime.datetime(1970, 1, 1, 0, 1, 20),
                 datetime.datetime(1970, 1, 1, 0, 1, 40)]
        exp = [20., 40., 60., 80, 100.]
        res = date2num(dates, self.unit, self.calendar)
        np.testing.assert_array_almost_equal(exp, res, decimal=4)

    def test_multidim_sequence(self):
        dates = [[datetime.datetime(1970, 1, 1, 0, 0, 20),
                  datetime.datetime(1970, 1, 1, 0, 0, 40),
                  datetime.datetime(1970, 1, 1, 0, 1)],
                 [datetime.datetime(1970, 1, 1, 0, 1, 20),
                  datetime.datetime(1970, 1, 1, 0, 1, 40),
                  datetime.datetime(1970, 1, 1, 0, 2)]]
        exp_shape = (2, 3)
        res = date2num(dates, self.unit, self.calendar)
        self.assertEqual(exp_shape, res.shape)

    def test_discard_mircosecond(self):
        date = datetime.datetime(1970, 1, 1, 0, 0, 5, 750000)
        exp = 5.
        res = date2num(date, self.unit, self.calendar)

        self.assertAlmostEqual(exp, res, places=4)

    def test_long_time_interval(self):
        # This test should fail with an error that we need to catch properly.
        unit = 'years since 1970-01-01'
        date = datetime.datetime(1970, 1, 1, 0, 0, 5)
        exp_emsg = 'interval of "months", "years" .* got "years".'
        with six.assertRaisesRegex(self, ValueError, exp_emsg):
            date2num(date, unit, self.calendar)


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2019, Met Office
#
# This file is part of cf-units.
#
# cf-units is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cf-units is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cf-units.  If not, see <http://www.gnu.org/licenses/>.

"""
Configure pytest to ignore python 3 files in python 2.

"""
import os.path
import glob
import six


if six.PY2:
    here = os.path.dirname(__file__)

    # Files under cf_units/_udunits2_parser/parser/ are all autogenerated
    # (and are py3 only).
    all_parse_py = glob.glob(
        os.path.join(here, '_udunits2_parser', 'parser', '*.py'))

    # Files under cf_units/_udunits2_parser are python3 *only*.
    all_compiled_parse_py = glob.glob(
        os.path.join(here, '_udunits2_parser', '*.py'))

    # Files under cf_units/tests/integration/parse are python3 *only*.
    parse_test_files = glob.glob(
        os.path.join(here, 'tests', 'integration', 'parse', '*.py'))

    # Files that are python3 only.
    python3_specific = [os.path.join(here, 'tex.py'),
                        os.path.join(here, 'tests', 'test_tex.py')]

    # collect_ignore is the special variable that pytest reads to
    # indicate which files should be ignored (and not even imported).
    # See also https://docs.pytest.org/en/latest/example/pythoncollection.html
    collect_ignore = (list(all_parse_py) +
                      list(all_compiled_parse_py) +
                      list(parse_test_files) +
                      python3_specific)



# (C) British Crown Copyright 2017 - 2018, Met Office
#
# This file is part of cf-units.
#
# cf-units is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cf-units is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cf-units.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for the `cf_units._udunits2` module.

In most cases, we don't test the correctness
of the operations, only that they return valid objects or raise an
exception where expected."""

from __future__ import (absolute_import, division, print_function)
from six.moves import (filter, input, map, range, zip)  # noqa

import errno
import unittest

import numpy as np

from cf_units import _udunits2 as _ud


_ud.set_error_message_handler(_ud.ignore)


class Test_get_system(unittest.TestCase):
    """
    Test case for operations which create a system object.

    """
    def test_read_xml(self):
        system = _ud.read_xml()

        self.assertIsNotNone(system)

    def test_read_xml_invalid_path(self):
        with self.assertRaises(_ud.UdunitsError) as cm:
            _ud.read_xml(b'/not/a/path.xml')
        ex = cm.exception

        self.assertEqual(ex.status_msg(), 'UT_OPEN_ARG')
        self.assertEqual(ex.errnum, errno.ENOENT)


class Test_system(unittest.TestCase):
    """
    Test case for system operations.

    """
    def setUp(self):
        self.system = _ud.read_xml()

    def test_get_unit_by_name(self):
        unit = _ud.get_unit_by_name(self.system, b'metre')

        self.assertIsNotNone(unit)

    def test_get_unit_by_name_invalid_unit(self):
        with self.assertRaises(_ud.UdunitsError):
            _ud.get_unit_by_name(self.system, b'jigawatt')

    def test_parse(self):
        unit = _ud.parse(self.system, b'gigawatt', _ud.UT_ASCII)

        self.assertIsNotNone(unit)

    def test_parse_latin1(self):
        angstrom = _ud.parse(self.system, b'\xe5ngstr\xF6m', _ud.UT_LATIN1)

        self.assertIsNotNone(angstrom)

    def test_parse_ISO_8859_1(self):
        angstrom = _ud.parse(self.system, b'\xe5ngstr\xF6m', _ud.UT_ISO_8859_1)

        self.assertIsNotNone(angstrom)

    def test_parse_UTF8(self):
        angstrom = _ud.parse(self.system, b'\xc3\xa5ngstr\xc3\xb6m',
                             _ud.UT_UTF8)

        self.assertIsNotNone(angstrom)

    def test_parse_invalid_unit(self):
        with self.assertRaises(_ud.UdunitsError):
            _ud.parse(self.system, b'jigawatt', _ud.UT_ASCII)


class Test_unit(unittest.TestCase):
    """
    Test case for unit operations.

    """
    def setUp(self):
        self.system = _ud.read_xml()
        self.metre = _ud.get_unit_by_name(self.system, b'metre')
        self.yard = _ud.get_unit_by_name(self.system, b'yard')
        self.second = _ud.get_unit_by_name(self.system, b'second')

    def test_clone(self):
        metre_clone = _ud.clone(self.metre)

        self.assertIsNot(metre_clone, self.metre)

    def test_is_dimensionless_true(self):
        radian = _ud.get_unit_by_name(self.system, b'radian')
        self.assertTrue(_ud.is_dimensionless(radian))

    def test_is_dimensionless_false(self):
        self.assertFalse(_ud.is_dimensionless(self.metre))

    def test_compare_same_unit(self):
        self.assertEqual(_ud.compare(self.metre, self.metre), 0)

    def test_compare_diff_unit(self):
        comp = _ud.compare(self.metre, self.second)
        comp_r = _ud.compare(self.second, self.metre)

        self.assertNotEqual(comp, 0)
        self.assertNotEqual(comp_r, 0)
        # m < s iff s > m
        self.assertEqual(comp < 0, comp_r > 0)

    def test_are_convertible_true(self):
        self.assertTrue(_ud.are_convertible(self.metre, self.yard))

    def test_are_convertible_false(self):
        self.assertFalse(_ud.are_convertible(self.metre, self.second))

    def test_get_converter_valid(self):
        _ud.get_converter(self.metre, self.yard)

    def test_get_converter_invalid(self):
        with self.assertRaises(_ud.UdunitsError) as cm:
            _ud.get_converter(self.metre, self.second)
        ex = cm.exception

        self.assertEqual(ex.status_msg(), 'UT_MEANINGLESS')

    def test_scale(self):
        mm = _ud.scale(0.001, self.metre)

        self.assertIsNotNone(mm)

    def test_offset(self):
        kelvin = _ud.get_unit_by_name(self.system, b'kelvin')
        celsius = _ud.offset(kelvin, 273.15)

        self.assertIsNotNone(celsius)

    def test_offset_by_time_valid(self):
        time_since = _ud.offset_by_time(self.second, -31622400.0)

        self.assertIsNotNone(time_since)

    def test_offset_by_time_invalid(self):
        with self.assertRaises(_ud.UdunitsError) as cm:
            _ud.offset_by_time(self.metre, -31622400.0)
        cm.exception

        # The udunits package should set a status of UT_MEANINGLESS, according
        # to the documentation. However, it is setting it to UT_SUCCESS.
        # self.assertEqual(ex.status_msg(), 'UT_MEANINGLESS')

    def test_multiply(self):
        metre_second = _ud.multiply(self.metre, self.second)

        self.assertIsNotNone(metre_second)

    def test_invert(self):
        hertz = _ud.invert(self.second)

        self.assertIsNotNone(hertz)

    def test_divide(self):
        metres_per_second = _ud.divide(self.metre, self.second)

        self.assertIsNotNone(metres_per_second)

    def test_raise_(self):
        sq_metre = _ud.raise_(self.metre, 2)

        self.assertIsNotNone(sq_metre)

    def test_root(self):
        hectare = _ud.get_unit_by_name(self.system, b'hectare')
        hundred_metre = _ud.root(hectare, 2)

        self.assertIsNotNone(hundred_metre)

    def test_log(self):
        log_metre = _ud.log(2.7182818, self.metre)

        self.assertIsNotNone(log_metre)

    def test_format(self):
        pascal = _ud.get_unit_by_name(self.system, b'pascal')
        symb = _ud.format(pascal)
        name = _ud.format(pascal, _ud.UT_NAMES)
        defn = _ud.format(pascal, _ud.UT_DEFINITION)
        name_defn = _ud.format(pascal, _ud.UT_DEFINITION | _ud.UT_NAMES)

        self.assertEqual(symb, b'Pa')
        self.assertEqual(name, b'pascal')
        self.assertEqual(defn, b'm-1.kg.s-2')
        self.assertEqual(name_defn, b'meter^-1-kilogram-second^-2')


class Test_time_encoding(unittest.TestCase):

    def setUp(self):
        self.year, self.month, self.day = 2000, 1, 1
        self.date_encoding = -31622400.0
        self.hours, self.minutes, self.seconds = 12, 34, 56
        self.clock_encoding = 45296.0

    def test_encode_date(self):
        res_date_encoding = _ud.encode_date(self.year, self.month, self.day)

        self.assertEqual(self.date_encoding, res_date_encoding)

    def test_encode_clock(self):
        res_clock_encoding = _ud.encode_clock(self.hours, self.minutes,
                                              self.seconds)

        self.assertEqual(self.clock_encoding, res_clock_encoding)

    def test_encode_time(self):
        res_time_encoding = _ud.encode_time(self.year, self.month, self.day,
                                            self.hours, self.minutes,
                                            self.seconds)

        self.assertEqual(self.clock_encoding + self.date_encoding,
                         res_time_encoding)

    def test_decode_time(self):
        res_year, res_month, res_day, res_hours, res_minutes, res_seconds,\
            res_resolution =\
            _ud.decode_time(self.date_encoding + self.clock_encoding)

        self.assertEqual(
            (res_year, res_month, res_day, res_hours, res_minutes),
            (self.year, self.month, self.day, self.hours, self.minutes))
        self.assertTrue(res_seconds - res_resolution < self.seconds <
                        res_seconds + res_resolution)


class Test_convert(unittest.TestCase):
    """
    Test case for convert operations.

    """
    def setUp(self):
        system = _ud.read_xml()
        metre = _ud.get_unit_by_name(system, b'metre')
        yard = _ud.get_unit_by_name(system, b'yard')
        self.converter = _ud.get_converter(metre, yard)
        self.factor = 1.0936132669448853

    def test_convert_float(self):
        res = _ud.convert_float(self.converter, 2.5)
        np.testing.assert_approx_equal(2.5 * self.factor, res)

    def test_convert_floats(self):
        arr = np.array([2.5, 5., 10.], dtype=np.float32)
        res = np.empty_like(arr)
        _ud.convert_floats(self.converter, arr, res)
        np.testing.assert_array_almost_equal(arr * self.factor, res)

    def test_convert_double(self):
        res = _ud.convert_double(self.converter, 2.5)
        np.testing.assert_approx_equal(2.5 * self.factor, res)

    def test_convert_doubles(self):
        arr = np.array([2.5, 5., 10.], dtype=np.float64)
        res = np.empty_like(arr)
        _ud.convert_doubles(self.converter, arr, res)
        np.testing.assert_array_almost_equal(arr * self.factor, res)


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2015 - 2019, Met Office
#
# This file is part of cf-units.
#
# cf-units is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cf-units is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cf-units.  If not, see <http://www.gnu.org/licenses/>.

"""
Compiles the UDUNITS-2 grammar using ANTLR4.

You may be interested in running this with entr to watch changes to the
grammar:

    echo udunits2*.g4* | entr -c "python compile.py"

You're welcome ;).

"""

import collections
from pathlib import Path
import re
import subprocess
import urllib.request

try:
    import jinja2
except ImportError:
    raise ImportError('Jinja2 needed to compile the grammar.')


JAR_NAME = 'antlr-4.7.2-complete.jar'
JAR_URL = f'https://www.antlr.org/download/{JAR_NAME}'
HERE = Path(__file__).resolve().parent

JAR = HERE / JAR_NAME

LEXER = HERE / 'parser' / 'udunits2Lexer.g4'
PARSER = HERE / 'udunits2Parser.g4'


def expand_lexer(source, target):
    MODE_P = re.compile(r'mode ([A-Z_]+)\;')
    TOKEN_P = re.compile(r'([A-Z_]+) ?\:.*')

    with open(source, 'r') as fh:
        content = fh.read()

    template = jinja2.Environment(
        loader=jinja2.BaseLoader).from_string(content)

    current_mode = 'DEFAULT_MODE'

    tokens = collections.defaultdict(list)

    for line in content.split('\n'):
        mode_g = MODE_P.match(line)
        if mode_g:
            current_mode = mode_g.group(1)

        token_g = TOKEN_P.match(line)
        if token_g:
            tokens[current_mode].append(token_g.group(1))

    new_content = template.render(tokens=tokens)
    with open(target, 'w') as fh:
        fh.write(new_content)


def main():
    if not JAR.exists():
        print(f'Downloading {JAR_NAME}...')
        urllib.request.urlretrieve(JAR_URL, str(JAR))

    print('Expanding lexer...')
    expand_lexer(LEXER.parent.parent / (LEXER.name + '.jinja'), str(LEXER))

    print('Compiling lexer...')
    subprocess.run(
        ['java', '-jar', str(JAR), '-Dlanguage=Python3',
         str(LEXER), '-o', 'parser'],
        check=True)

    print('Compiling parser...')
    subprocess.run(
        ['java', '-jar', str(JAR), '-Dlanguage=Python3',
         '-no-listener', '-visitor',
         str(PARSER), '-o', 'parser'],
        check=True)

    print('Done.')


if __name__ == '__main__':
    main()



# (C) British Crown Copyright 2019, Met Office
#
# This file is part of cf-units.
#
# cf-units is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cf-units is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cf-units.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import six

from cf_units.tex import tex


if six.PY2:
    pytest.skip('skipping latex in Python2')


def test_basic():
    u = 'kg kg-1'
    assert tex(u) == r'{kg}\cdot{{kg}^{-1}}'


def test_identifier_micro():
    u = 'microW m-2'
    assert tex(u) == r'{{\mu}W}\cdot{{m}^{-2}}'


def test_raise():
    u = 'm^2'
    assert tex(u) == r'{m}^{2}'


def test_multiply():
    u = 'foo bar'
    assert tex(u) == r'{foo}\cdot{bar}'


def test_divide():
    u = 'foo per bar'
    assert tex(u) == r'\frac{foo}{bar}'


def test_shift():
    u = 'foo @ 50'
    assert tex(u) == r'{foo} @ {50}'


def test_complex():
    u = 'microW^2 / (5^-2)π per sec @ 42'
    expected = r'{\frac{{\frac{{{\mu}W}^{2}}{{5}^{-2}}}\cdot{π}}{sec}} @ {42}'
    assert tex(u) == expected



# (C) British Crown Copyright 2016 - 2018, Met Office
#
# This file is part of cf-units.
#
# cf-units is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cf-units is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cf-units.  If not, see <http://www.gnu.org/licenses/>.
"""Test function :func:`cf_units._num2date_to_nearest_second`."""

from __future__ import (absolute_import, division, print_function)
from six.moves import (filter, input, map, range, zip)  # noqa

import unittest
import datetime

import numpy as np
import cftime

from cf_units import _num2date_to_nearest_second


class Test(unittest.TestCase):
    def setup_units(self, calendar):
        self.useconds = cftime.utime('seconds since 1970-01-01',  calendar)
        self.uminutes = cftime.utime('minutes since 1970-01-01', calendar)
        self.uhours = cftime.utime('hours since 1970-01-01', calendar)
        self.udays = cftime.utime('days since 1970-01-01', calendar)

    def check_dates(self, nums, utimes, expected):
        for num, utime, exp in zip(nums, utimes, expected):
            res = _num2date_to_nearest_second(num, utime)
            self.assertEqual(exp, res)

    def check_timedelta(self, nums, utimes, expected):
        for num, utime, exp in zip(nums, utimes, expected):
            epoch = utime.num2date(0)
            res = _num2date_to_nearest_second(num, utime)
            delta = res - epoch
            seconds = np.round(delta.seconds + (delta.microseconds / 1000000))
            self.assertEqual((delta.days, seconds), exp)

    def test_scalar(self):
        utime = cftime.utime('seconds since 1970-01-01',  'gregorian')
        num = 5.
        exp = datetime.datetime(1970, 1, 1, 0, 0, 5)
        res = _num2date_to_nearest_second(num, utime)
        self.assertEqual(exp, res)

    def test_sequence(self):
        utime = cftime.utime('seconds since 1970-01-01',  'gregorian')
        nums = [20., 40., 60., 80, 100.]
        exp = [datetime.datetime(1970, 1, 1, 0, 0, 20),
               datetime.datetime(1970, 1, 1, 0, 0, 40),
               datetime.datetime(1970, 1, 1, 0, 1),
               datetime.datetime(1970, 1, 1, 0, 1, 20),
               datetime.datetime(1970, 1, 1, 0, 1, 40)]
        res = _num2date_to_nearest_second(nums, utime)
        np.testing.assert_array_equal(exp, res)

    def test_multidim_sequence(self):
        utime = cftime.utime('seconds since 1970-01-01',  'gregorian')
        nums = [[20., 40., 60.],
                [80, 100., 120.]]
        exp_shape = (2, 3)
        res = _num2date_to_nearest_second(nums, utime)
        self.assertEqual(exp_shape, res.shape)

    def test_masked_ndarray(self):
        utime = cftime.utime('seconds since 1970-01-01',  'gregorian')
        nums = np.ma.masked_array([20., 40., 60.], [False, True, False])
        exp = [datetime.datetime(1970, 1, 1, 0, 0, 20),
               None,
               datetime.datetime(1970, 1, 1, 0, 1)]
        res = _num2date_to_nearest_second(nums, utime)
        np.testing.assert_array_equal(exp, res)

    # Gregorian Calendar tests

    def test_simple_gregorian(self):
        self.setup_units('gregorian')
        nums = [20., 40.,
                75., 150.,
                8., 16.,
                300., 600.]
        utimes = [self.useconds, self.useconds,
                  self.uminutes, self.uminutes,
                  self.uhours, self.uhours,
                  self.udays, self.udays]
        expected = [datetime.datetime(1970, 1, 1, 0, 0, 20),
                    datetime.datetime(1970, 1, 1, 0, 0, 40),
                    datetime.datetime(1970, 1, 1, 1, 15),
                    datetime.datetime(1970, 1, 1, 2, 30),
                    datetime.datetime(1970, 1, 1, 8),
                    datetime.datetime(1970, 1, 1, 16),
                    datetime.datetime(1970, 10, 28),
                    datetime.datetime(1971, 8, 24)]

        self.check_dates(nums, utimes, expected)

    def test_fractional_gregorian(self):
        self.setup_units('gregorian')
        nums = [5./60., 10./60.,
                15./60., 30./60.,
                8./24., 16./24.]
        utimes = [self.uminutes, self.uminutes,
                  self.uhours, self.uhours,
                  self.udays, self.udays]
        expected = [datetime.datetime(1970, 1, 1, 0, 0, 5),
                    datetime.datetime(1970, 1, 1, 0, 0, 10),
                    datetime.datetime(1970, 1, 1, 0, 15),
                    datetime.datetime(1970, 1, 1, 0, 30),
                    datetime.datetime(1970, 1, 1, 8),
                    datetime.datetime(1970, 1, 1, 16)]

        self.check_dates(nums, utimes, expected)

    def test_fractional_second_gregorian(self):
        self.setup_units('gregorian')
        nums = [0.25, 0.5, 0.75,
                1.5, 2.5, 3.5, 4.5]
        utimes = [self.useconds] * 7
        expected = [datetime.datetime(1970, 1, 1, 0, 0, 0),
                    datetime.datetime(1970, 1, 1, 0, 0, 1),
                    datetime.datetime(1970, 1, 1, 0, 0, 1),
                    datetime.datetime(1970, 1, 1, 0, 0, 2),
                    datetime.datetime(1970, 1, 1, 0, 0, 3),
                    datetime.datetime(1970, 1, 1, 0, 0, 4),
                    datetime.datetime(1970, 1, 1, 0, 0, 5)]

        self.check_dates(nums, utimes, expected)

    # 360 day Calendar tests

    def test_simple_360_day(self):
        self.setup_units('360_day')
        nums = [20., 40.,
                75., 150.,
                8., 16.,
                300., 600.]
        utimes = [self.useconds, self.useconds,
                  self.uminutes, self.uminutes,
                  self.uhours, self.uhours,
                  self.udays, self.udays]
        # Expected results in (days, seconds) delta from utime epoch.
        expected = [(0, nums[0]),
                    (0, nums[1]),
                    (0, nums[2]*60),
                    (0, nums[3]*60),
                    (0, nums[4]*60*60),
                    (0, nums[5]*60*60),
                    (nums[6], 0),
                    (nums[7], 0)]

        self.check_timedelta(nums, utimes, expected)

    def test_fractional_360_day(self):
        self.setup_units('360_day')
        nums = [5./60., 10./60.,
                15./60., 30./60.,
                8./24., 16./24.]
        utimes = [self.uminutes, self.uminutes,
                  self.uhours, self.uhours,
                  self.udays, self.udays]
        # Expected results in (days, seconds) delta from utime epoch.
        expected = [(0, nums[0]*60),
                    (0, nums[1]*60),
                    (0, nums[2]*60*60),
                    (0, nums[3]*60*60),
                    (0, nums[4]*24*60*60),
                    (0, nums[5]*24*60*60)]

        self.check_timedelta(nums, utimes, expected)

    def test_fractional_second_360_day(self):
        self.setup_units('360_day')
        nums = [0.25, 0.5, 0.75,
                1.5, 2.5, 3.5, 4.5]
        utimes = [self.useconds] * 7
        # Expected results in (days, seconds) delta from utime epoch.
        expected = [(0, 0.0),  # rounded down to this
                    (0, 1.0),  # rounded up to this
                    (0, 1.0),  # rounded up to this
                    (0, 2.0),  # rounded up to this
                    (0, 3.0),  # rounded up to this
                    (0, 4.0),  # rounded up to this
                    (0, 5.0)]  # rounded up to this

        self.check_timedelta(nums, utimes, expected)

    # 365 day Calendar tests

    def test_simple_365_day(self):
        self.setup_units('365_day')
        nums = [20., 40.,
                75., 150.,
                8., 16.,
                300., 600.]
        utimes = [self.useconds, self.useconds,
                  self.uminutes, self.uminutes,
                  self.uhours, self.uhours,
                  self.udays, self.udays]
        # Expected results in (days, seconds) delta from utime epoch.
        expected = [(0, nums[0]),
                    (0, nums[1]),
                    (0, nums[2]*60),
                    (0, nums[3]*60),
                    (0, nums[4]*60*60),
                    (0, nums[5]*60*60),
                    (nums[6], 0),
                    (nums[7], 0)]

        self.check_timedelta(nums, utimes, expected)

    def test_fractional_365_day(self):
        self.setup_units('365_day')
        nums = [5./60., 10./60.,
                15./60., 30./60.,
                8./24., 16./24.]
        utimes = [self.uminutes, self.uminutes,
                  self.uhours, self.uhours,
                  self.udays, self.udays]
        # Expected results in (days, seconds) delta from utime epoch.
        expected = [(0, nums[0]*60),
                    (0, nums[1]*60),
                    (0, nums[2]*60*60),
                    (0, nums[3]*60*60),
                    (0, nums[4]*24*60*60),
                    (0, nums[5]*24*60*60)]

        self.check_timedelta(nums, utimes, expected)

    def test_fractional_second_365_day(self):
        self.setup_units('365_day')
        nums = [0.25, 0.5, 0.75,
                1.5, 2.5, 3.5, 4.5]
        utimes = [self.useconds] * 7
        # Expected results in (days, seconds) delta from utime epoch.
        expected = [(0, 0.0),  # rounded down to this
                    (0, 1.0),  # rounded up to this
                    (0, 1.0),  # rounded up to this
                    (0, 2.0),  # rounded up to this
                    (0, 3.0),  # rounded up to this
                    (0, 4.0),  # rounded up to this
                    (0, 5.0)]  # rounded up to this

        self.check_timedelta(nums, utimes, expected)


if __name__ == '__main__':
    unittest.main()
